' ###############################################################################################
' ScraperScriptClass  -- Class for processing a text file scrape.
' ###############################################################################################
' Version 1.0, 23 Jul 2005
' Author: John J. Schultz
'
' Version History:
'   1.0 -- original version
' ###################################################################################

Class ScraperScriptClass
	'Const ForReading = 1, ForWriting = 2, ForAppending = 8

	Dim iTokenCount
	Dim asTokens ()   ' Name, Active (True, False), Value

	' Properties
	Dim iProcessCode
	Dim sProcessMessage
	Dim sExtractedContent
	Dim iListCount
	Dim aList()
	Dim iListValueCount
	Dim aListValue()

	Private Sub Class_Initialize
  		iProcessCode = 0
  		sProcessMessage = "No processing done"
  		sExtractedContent = ""
		iListCount = 0
		iListValueCount = 0
	End Sub

	Property Get ProcessCode
	  Set ProcessCode = iProcessCode
	End Property

	Property Get ProcessMessage
		Set ProcessMessage = sProcessMessage
	End Property

	Property Get sExtractedContent
		Set ExtractedContent = sExtractedContent
	End Property

	Property Get ListCount
		Set ExtractedContent = iListCount
	End Property

	Property Get List
		Set ExtractedContent = aList
	End Property

	Property Get ListValueCount
		Set ExtractedContent = iListValueCount
	End Property

	Property Get ListValue
		Set ExtractedContent = aListValue
	End Property

	Function MakeTimeStamp()
		Dim dNow

		dNow = Now()
		MakeTimeStamp = _
			right ("000" & Year(dNow), 4) & _
			right ("0" & Month(dNow), 2) & _
			right ("0" & Day(dNow), 2) & _
			"_" & _
			right ("0" & Hour(dNow), 2) & _
			right ("0" & Minute(dNow), 2) & _
			right ("0" & Second(dNow), 2)
	End Function

	Function FileExists (sFileName)
		Dim oFSO

		Set oFSO = CreateObject("Scripting.FileSystemObject")
		FileExists = oFSO.FileExists(sFileName)
		set oFSO = Nothing
	End Function

	Function FileToString (sFileName, iTristate)
		Dim oFSO, oTS

		Set oFSO = CreateObject("Scripting.FileSystemObject")
		Set oTS = oFSO.OpenTextFile(sFileName, 1, True, iTristate)

		FileToString = oTS.ReadAll
		oTS.Close()

		set oTS = Nothing
		set oFSO = Nothing
	End Function

	Sub StringToFile (sFileName, sText, iTristate)
		Dim oFSO, oTS

		Set oFSO = CreateObject("Scripting.FileSystemObject")
		Set oTS = oFSO.OpenTextFile(sFileName, 2, True, iTristate)

		oTS.Write sText
		oTS.Close()

		set oTS = Nothing
		set oFSO = Nothing
	End Sub

	Function LoadEnvironmentVariable (sToken)
		' Recursively resolves any tokens embedded in the return value so that the actual value comes back.
		' These are normally certain environment variables coming from the PROCESS section: USERPROFILE, SYSTEMROOT
		' but anything is replaced.
		Dim WshUserEnv, iIndex1, iIndex2, sBase, sValue

		set WshUserEnv = WshShell.Environment(sSource)
		sBase = WshUserEnv(sToken)

		iIndex1 = Instr (1, sBase, "%")
		If iIndex1 > 0 Then
  			iIndex2 = Instr(1, sBase, "%USERPROFILE%")
  			If iIndex2 = iIndex1 Then
  				sBase = ReplaceNoCase (sBase, "%USERPROFILE%", LoadEnvironmentVariable ("USERPROFILE", "PROCESS"))
  			End If
  			iIndex2 = Instr(1, UCASE (sBase), "%SYSTEMROOT%")
  			If iIndex2 = iIndex1 Then
  				sBase = ReplaceNoCase (sBase, "%SYSTEMROOT%", LoadEnvironmentVariable ("SYSTEMROOT", "PROCESS"))
  			End If
  			iIndex2 = Instr (iIndex1 + 1, sBase, "%")
  			If iIndex2 > 0 Then
  				sValue = LoadEnvironmentVariable (Mid (sBase, iIndex1 + 1, iIndex2 - iIndex1 - 1), sSource)
  				sBase = REPLACE (sBase, Mid (sBase, iIndex1, iIndex2 - iIndex1 + 1), sValue)
  			End If
		End If
		set WshUserEnv = Nothing
		LoadEnvironmentVariable = sBase
	End Function

	Function RegExFramedFind (sRawTest, sStartPattern, bStartIgnoreCase, iStartIteration, sEndPattern, bEndIgnoreCase, iEndIteration)
		' Returns an array of strings, containing the blocks matched.
		' of the subset.
		'  E.g.:
		' BlockA,Start,1,2,3,End,BlockB,Start,4,5,6,End,BlockC,Start,7,8,9,End
		' StartIteration = 0: EndIteration = 0: StartPattern = "BlockA," : EndPattern ="End" : Returns BlockA,Start,1,2,3,End
		' Iteration = 0: EndIteration = 0: StartPattern = "BlockB," : EndPattern ="End" : Returns BlockB,Start,4,5,6,End
		' Iteration = 2: EndIteration = 0: StartPattern = "Block.," : EndPattern ="End" : Returns BlockB,Start,4,5,6,End
		' etc.
		Dim oRegEx
		Dim oMatch
		Dim oMatches
		Dim oInMatch
		Dim oInMatches
		Dim sRetStr
		Dim iOutIndex
		Dim iInIndex
		Dim iFrameCount

		iFrameCount = 0
		Set oRegEx = New RegExp   ' Create regular expression.
		With oRegEx
			.Pattern = sStartPattern   ' Set pattern.
			.IgnoreCase = bStartIgnoreCase   ' Set case insensitivity.
			If iStartIteration = 1 Then
				.Global = True   ' when only match #1 is required.
			Else
				.Global = True   ' Set global applicability.
			End If
	   		Set oMatches = .Execute(sRawTest)   ' Execute search.
		End With
		sRetStr = ""
		iOutIndex = 0
		For Each oMatch in oMatches   ' Iterate Matches collection.
			iOutIndex = iOutIndex + 1
			If iStartIteration = 0 Or iStartIteration = iOutIndex Then
				With oMatch
					With oRegEx
						.Pattern = sEndPattern   ' Set pattern.
						.IgnoreCase = bEndIgnoreCase   ' Set case insensitivity.
						If iEndIteration = 1 Then
							.Global = True   ' when only match #1 is required.
						Else
							.Global = True   ' Set global applicability.
						End If
	   					Set oInMatches = .Execute(Mid (sRawTest, oMatch.FirstIndex))   ' Execute search.
					End With
					iInIndex = 0
					For Each oInMatch in oInMatches   ' Iterate Matches collection.
						iInIndex = iInIndex + 1
						If iEndIteration = 0 Or iEndIteration = iInIndex Then
							If sRetStr <> "" Then
								sRetStr = sRetStr & ","
							End If
							sRetStr = sRetStr & (oMatch.FirstIndex + 1) & "," & (oInMatch.FirstIndex + oInMatch.Length - 1)
							iFrameCount = iFrameCount + 1
						End If
					Next
				End With
			End If
		Next

		Set oInMatch = Nothing
		Set oInMatches = Nothing
		Set oMatch = Nothing
		Set oMatches = Nothing
		Set oRegEx = Nothing

		If sRetStr <> "" Then
			sRetStr = "," & sRetStr
		End If
		sRetStr = iFrameCount & sRetStr

		RegExFramedFind = sRetStr
	End Function

	Function GetColumnFromString (sString, sDelimiter, iPosition)
		Dim aResults

		aResults = Split (sString, sDelimiter, -1, 1)
		If iPosition <= UBound (aResults) Then
			GetColumnFromString = aResults (iPosition)
		Else
			GetColumnFromString = ""
		End If
	End Function

	Function RegExFind (sPattern, sRawTest, bIgnoreCase, iIteration)
		Dim oRegEx, oMatch, oMatches, iIndex, iCount, aMatches()   ' Create variable.

		Set oRegEx = New RegExp   ' Create regular expression.
		With oRegEx
			.Pattern = sPattern   ' Set pattern.
			.IgnoreCase = bIgnoreCase   ' Set case insensitivity.
			If iIteration = 1 Then
				.Global = True   ' when only match #1 is required.
			Else
				.Global = True   ' Set global applicability.
			End If
	   		Set oMatches = .Execute(sRawTest)   ' Execute search.
		End With
		sRetStr = ""
		iIndex = 0
		iCount = -1
		For Each oMatch in oMatches   ' Iterate Matches collection.
			iIndex = iIndex + 1
			If iIteration = 0 Or iIteration = iIndex Then
		    iCount = iCount + 1
		    ReDim Preserve aMatches (iCount)
				aMatches (iCount) = oMatch.Value
			End If
		Next
		Set oMatch = Nothing
		Set oMatches = Nothing
		Set oRegEx = Nothing
		RegExFind = aMatches
	End Function

	Function ReplaceNoCase (strSource, strFind, strReplace)   ' Case-insensitive equivalent of Replace()
		Dim iIndex, sBase, sValue

		sBase = strSource
		iIndex = Instr(1, UCASE (sBase), UCase (strFind))
		While iIndex > 0
			If iIndex + Len (strFind) > Len (sBase) Then
				sValue = left (sBase, iIndex - 1) & strReplace
			Else
				sValue = left (sBase, iIndex - 1) & strReplace & mid (sBase, iIndex + Len(strFind))
			End If
			sBase = sValue
			iIndex = Instr(1, UCASE (sBase), UCase (strFind))
		WEnd
		ReplaceNoCase = sBase
	End Function

	Function ReplaceTokens (psSource, psMatchPattern, psNamePattern)
		Dim iTokenIndex, sWork, sTokenName, sTokenValue, oRegEx, oMatches
		Dim sSource

		sSource = psSource

		Set oRegEx = New RegExp   ' Create regular expression.
		Do While True

			With oRegEx
				.Pattern = psMatchPattern   ' Set pattern.
				.IgnoreCase = False   ' Set case insensitivity.
				.Global = False   ' when only match #1 is required.
		   		Set oMatches = .Execute(sSource)   ' Execute search.
			End With

			If oMatches.Count = 0 Then
				Exit Do
			End If

			sWork = oMatches(0).Value

			With oRegEx
				.Pattern = psNamePattern   ' Set pattern.
				.IgnoreCase = False   ' Set case insensitivity.
				.Global = False   ' when only match #1 is required.
		   		Set oMatches = .Execute(sWork)   ' Execute search.
			End With

			sTokenName = ""
			If oMatches.Count = 1 Then
				sTokenName = oMatches(0).Value
			End If

			If sTokenName <> "" Then
				sTokenValue = GetTokenValue (sTokenName, "")
			Else
				sTokenValue = ""
			End If

			sSource = Replace (sSource, sWork, sTokenValue)
		Loop

		Set oMatches = Nothing
		Set oRegEx = Nothing

		ReplaceTokens = sSource
	End Function

	Function TokenIsValid (sTokenName, sTokenNamePattern)
		' Returns True if the Token name is valid, or False if not
		Dim oRegEx, oMatches, bResult
		Dim sSource

		Set oRegEx = New RegExp   ' Create regular expression.
		With oRegEx
			.Pattern = sTokenNamePattern   ' Set pattern.
			.IgnoreCase = True   ' Set case insensitivity.
			.Global = True   ' when only match #1 is required.
		   	Set oMatches = .Execute(sTokenName)   ' Execute search.
		End With

		TokenIsValid = (oMatches.Count = 1)
		Set oMatches = Nothing
		Set oRegEx = Nothing
	End Function

	Function TokenExists (sName)
		' Add and Edit combined
		Dim iIndex

		For iIndex = 0 to iTokenCount - 1
			If asTokens (0, iIndex) = UCase (sName) And asTokens (1, iIndex) Then
				Exit For
			End If
		Next

		TokenExists = (iIndex < iTokenCount)
	End Function

	Sub ListAllTokens ()
		' Lists all token values with Echo.  For debug purposes only
		Dim iIndex

		For iIndex = 0 to iTokenCount - 1
			If asTokens (1, iIndex) Then
				WScript.Echo asTokens (0, iIndex) & ", " & Mid ("YN", 2 + CBool (asTokens (1, iIndex) = True), 1) & ", " & asTokens (2, iIndex) ' & vbCrLf
			End If
		Next

	End Sub

	Function GetTokenValue (sName, sDefault)
		' Returns empty string if doesn't exist.
		Dim iIndex

		For iIndex = 0 to iTokenCount - 1
			If asTokens (0, iIndex) = UCase (sName) And asTokens (1, iIndex) Then
				Exit For
			End If
		Next

		If iIndex < iTokenCount Then
			GetTokenValue = asTokens (2, iIndex)
		Else
			GetTokenValue = sDefault
		End If
	End Function

	Function StoreToken (sName, sValue)
		' Add and Edit combined
		Dim iIndex

		For iIndex = 0 to iTokenCount - 1
			If asTokens (0, iIndex) = UCase (sName) And asTokens (1, iIndex) Then
				exit for
			End If
		Next

		If iIndex = iTokenCount Then
			iIndex =  iTokenCount
			ReDim Preserve asTokens (3, iTokenCount)
			iTokenCount = iTokenCount + 1
		End If

		If iIndex <= UBound (asTokens, 2) Then
			asTokens (0, iIndex) = UCASE(sName)
			asTokens (1, iIndex) = True
			asTokens (2, iIndex) = sValue
		End If
		StoreToken = (iIndex <= UBound (asTokens, 2))
	End Function

	Sub DropToken (sName)   ' Includes properties
		' Add and Edit combined
		Dim iIndex

		For iIndex = 0 to iTokenCount - 1
			If asTokens (1, iIndex) Then
				If asTokens (0, iIndex) = UCase (sName) Or Left (asTokens (0, iIndex) & Space (Len (sName) + 1), Len (sName) + 1) = UCase (sName) & "." Or Left (asTokens (0, iIndex) & Space (Len (sName) + 1), Len (sName) + 1) = UCase (sName) & "(" Then
					asTokens (1, iIndex) = False
				End If
			End If
		Next
	End Sub

	Function ListIndex (sListName)
		If iListCount = 0 Then
			iResult = 0
		Else
			For i = 0 To UBound (aList, 2)
				If LCase (LTrim (RTrim (sListName))) = aList (0,i) Then
					Exit For
				End If
			Next
			If i <= UBound (aList, 2)) Then
				iResult = i
			Else
				iResult = iListCount
			End If
		End If
		CreateList = bResult
	End Function

	Function CreateList (sListName, bUnique)
		bResult = True
		If iListCount = 0 Then
			Dim aList(1,0)
		Else
			If ListIndex (sListName) < iListCount Then
				bResult = False   ' List already exists
			End If
			If bResult Then
				ReDim Preserve aList(1,i)
			End If
		End If
		If bResult Then
			aList(0,i) = LCase (LTrim (RTrim (sListName)))
			aList(1,i) = CBool (bUnique)
			iListCount = iListCount + 1
		End If
		CreateList = bResult
	End Function

	Function PushListValue (sListName, sValue)
		bResult = True
		If iListCount = 0 Then
			bResult = False
		Else
			i = ListIndex (sListName)
			If i = iListCount Then
				bResult = False   ' List does not exist
			End If
		End If
		
		' For Unique lists, ensure sValue does not already exist.
		If bResult Then
			If aList (1,i) And iListValueCount > 0 Then
				For i1 = 0 To UBound (aListValue, 2)
					If LCase (LTrim (RTrim (sListName))) = aListValue (0, i1) Then
						If aListValue (1, i1) And sValue = aListValue (2, i1) Then
							bResult = False   ' Unique value violation
							Exit For
						End If
					End If
				Next
			End If
		End If
		
		If bResult Then
			If iListValueCount = 0 Then
				Dim aListValue(2,0)
			Else
				ReDim Preserve aListValue(2,i1)
			End If
			aListValue(0,i1) = LCase (LTrim (RTrim (sListName)))
			aListValue(1,i1) = True
			aListValue(2,i1) = sValue
			iListValueCount = iListValueCount + 1
		End If
		PushListValue = bResult		
	End Function

	' Returns an array of all the values in the list.
	Function GetListValues (sListName)
		Dim aValues(), iValuesCount
		
		iValuesCount = 0
		i = ListIndex (sListName)
		If i < iListCount Then
			For i1 = 0 to iListValueCount - 1
				If aValueList (1, i1) And aValueList (0, i1) = LCase (LTrim (RTrim (sListName))) Then
					If iValuesCount = 0 Then
						Dim aValues (iValuesCount)
					Else
						ReDim Preserve aValues (iValuesCount)
					End If
					aValues (iValuesCount) = aValueList (2, i1)
					iValuesCount = iValuesCount + 1
				End If
			Next
		End If
		
		If iValuesCount > 0 Then
			GetListValues = aValues
		Else
			GetListValues = ""
		End If
	End Function

	' NOTE: ScrapeUsingScript is best used for simple no-POST, single page retrieval and scraping.
	' The above functions can be used within VbScript itself, to create a scraping action.

	Function ScrapeUsingScript (sScript, sContent)
		' Returns !!!ERROR!!! as the beginning sequence if error(s) were encountered.
		Dim sLine
		Dim iTokenIndex
		Dim sOutput
		Dim iArgCount
		Dim asArgs (20)
		Dim aiArgPos(20,2)
		Dim asLines (2000)
		Dim aiForNextLines(10,3)     ' 0 = line index, 1 = token name, 2 = increment, 3 = test
		Dim objXMLHTTP
		Dim sResponse
		Dim iErr
		Dim sErr
		Dim sErrPrefix
		Dim sErrors
		Dim sErrSuffix
		Dim sErrLine
		Dim iLineNum
		Dim iLineCount
		Dim iForNextDepth
		Dim iIfBlockLevel
		Dim iInIfBlockLevel
		Dim iIter
		Dim iTest
		Dim bResult
		Dim bTmp
		Dim bUsingReturn
		Dim sTmp
		Dim lTmp
		Dim lTmp1
		Dim lTmp2
		Dim xTmp
		Dim sTokenPattern
		Dim sNamePattern
		Dim iParmIndex
		Dim iParmLength
		Dim bFatal
		Dim oOutput

		Set oOutput = CreateObject ("ADODB.Stream")
		With oOutput
			.Open
			.Type = 2
			.Charset = "US-ASCII"
		End With

		sErrors = ""
		' Default patterns for token naming and framing
		sNamePattern = "([_A-Za-z](([A-Za-z0-9_])+)?\.[_A-Za-z](([A-Za-z0-9_])+)?|[_A-Za-z](([A-Za-z0-9_])+)?)"
		sTokenPattern = "\#\[\[" & sNamePattern & "\]\]\#"

		iTokenCount = 0

		' Set static constants as tokens
		sTmp = _
			"Line #[[__ERROR_LINENUM__]]#: #[[__ERROR_LINE__]]#" & vbCrLf & _
			"#[[__ERROR_NUM__]]#: #[[__ERROR_MSG__]]#" & vbCrLf

		Call StoreToken ("__ERROR_PREFIX__", "!!!ERROR!!!" & vbCrLf)
		Call StoreToken ("__ERROR_SUFFIX__", "")
		Call StoreToken ("__ERROR_BODY__", sTmp)
		Call StoreToken ("__ERRORS_ONLY__", "False")

		Call StoreToken ("__OUTPUT_HEADER__", "")
		Call StoreToken ("__OUTPUT_FOOTER__", "")

		xTmp = Now()
		Call StoreToken ("__DATESTAMP__", Year(xTmp) & Right ("0" & Month(xTmp), 2) & Right ("0" & Day(xTmp), 2))
		Call StoreToken ("__TIMESTAMP__", Right ("0" & Hour(xTmp), 2) & Right ("0" & Minute(xTmp), 2) & Right ("0" & Second(xTmp), 2))

		iLineCount = 0
		bUsingReturn = False
		Do While sScript <> ""
			If Instr (sScript, vbCrLf) > 0 Then
				sLine = LTrim (RTrim (LEFT (sScript, Instr (sScript, vbCrLf) - 1)))
				sScript = MID (sScript, Instr (sScript, vbCrLf) + 2)
			Else
				sLine = LTrim (RTrim (sScript))
				sScript = ""
			End If
			While Left (sLine & " ", 1) = vbTab
				sLine = Mid (sLine, 2)
			WEnd
			asLines (iLineCount) = sLine
			iLineCount = iLineCount + 1
		Loop

		sOutput = ""
		iLineNum = 0
		iForNextDepth = -1
		iIfBlockLevel = 0
		iInIfBlockLevel = 0

		bFatal = False
		Do While iLineNum < iLineCount And Not bFatal And Not bUsingReturn
			iErr = 0
			sErr = ""
			sLine = RegExFind ("\S(.)+", asLines (iLineNum), True, 1)(0)
			sErrLine = ""

			If sLine <> "" Then
				If InStr ("';#", Left (sLine, 1)) = 0 Then   ' Line is other than a comment
					For iArgCount = 0 to UBound (aiArgPos)
						 aiArgPos(iArgCount, 0) = 0
						 aiArgPos(iArgCount, 1) = 0
						 asArgs(iArgCount) = ""
					Next
					aiArgPos(0, 0) = 1
					iArgCount = -1
					While sLine <> "" AND iArgCount < UBound (aiArgPos)
						iArgCount = iArgCount + 1
						asArgs (iArgCount) = LTrim (RTrim (Left (sLine & " ", INSTR (sLine & " ", " "))))
						sLine = LTrim (RTrim (Mid (sLine & " ", INSTR (sLine & " ", " "))))
						aiArgPos (iArgCount, 1) = Len (asArgs (iArgCount))
						If iArgCount < UBound (aiArgPos) Then
							aiArgPos (iArgCount + 1, 0) = Len (asLines (iLineNum)) - Len (sLine) + 1
						End If
					WEnd

					asArgs (0) = UCase (asArgs (0))   ' command processed in upper case

					' -----------------------------------------------------------------------
					If asArgs (0) = "IF" Then
						If iArgCount >= 1 Then
							If iIfBlockLevel = iInIfBlockLevel Then
								sTmp = ReplaceTokens (Mid (asLines (iLineNum), aiArgPos (1, 0)), sTokenPattern, sNamePattern)
								iErr = 0
								On Error Resume Next
								If CBool (Eval (sTmp)) = -1 Then
									iInIfBlockLevel = iInIfBlockLevel + 1
								End If
								iErr = Err.Number
								sErr = Err.Description
								On Error Goto 0
								If iErr <> 0 Then
    							bFatal = True
								End If
							End If
							iIfBlockLevel = iIfBlockLevel + 1
						Else
							iErr = -1
							sErr = "Does not conform to prototype: IF {expression}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "ENDIF" Then
						If iArgCount = 0 Then
							iIfBlockLevel = iIfBlockLevel - 1
							If iInIfBlockLevel > iIfBlockLevel Then
								iInIfBlockLevel = iIfBlockLevel
							End If
							If iIfBlockLevel < 0 Then
								iErr = -1
								sErr = "orphan ENDIF: no matching IF statement"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: ENDIF"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf iIfBlockLevel <> iInIfBlockLevel Then
						' The command is skipped, since the IF/ENDIF block is preventing its execution.

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "TOKEN_MATCH_PATTERN" Then
						If iArgCount >= 1 Then
							sTokenPattern = Mid (asLines (iLineNum), aiArgPos (1, 0))
						Else
							iErr = -1
							sErr = "Does not conform to prototype: TOKEN_MATCH_PATTERN {regex}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "TOKEN_NAME_PATTERN" Then
						If iArgCount >= 1 Then
							sNamePattern = Mid (asLines (iLineNum), aiArgPos (1, 0))
						Else
							iErr = -1
							sErr = "Does not conform to prototype: TOKEN_NAME_PATTERN {regex}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "RETRIEVE" Then
						If iArgCount >= 1 Then
							sResponse = ReplaceTokens (Mid (asLines (iLineNum), aiArgPos (1, 0)), sTokenPattern, sNamePattern)

							Err.Clear
							On Error Resume Next
							Set objXMLHTTP = CreateObject("MSXML2.XMLHTTP")
							If Err.number = 0 Then objXMLHTTP.Open "GET", sResponse, False
							If Err.number = 0 Then objXMLHTTP.Send
							If Err.number = 0 Then
								If objXMLHTTP.Status = 200 Then
									sResponse = objXMLHTTP.ResponseText
									If Err.number <> 0 Then
									   iErr = Err.number
									End If
								Else
									iErr = -2
									sErr = "Unexpected server response: " & objXMLHTTP.Status
									bFatal = True
								End If
							End If
							If iErr > 0 Then
								iErr = Err.number
								sErr = Err.Description
								sErrLine = "RETRIEVE " & sResponse
								bFatal = True
							End If
							set objXMLHTTP = Nothing
							On Error Goto 0
							If iErr = 0 Then
								sContent = sResponse
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: RETRIEVE {url} [{POSTDATA}]"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "INCREMENT" Then
						If iArgCount = 1 Then
							If Not StoreToken (asArgs (1), CStr (Val ("0" & GetTokenValue (asArgs (1), "")) + 1)) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: INCREMENT {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "DECREMENT" Then
						If iArgCount = 1 Then
							If Not StoreToken (asArgs (1), CStr (Val ("0" & GetTokenValue (asArgs (1), "")) - 1)) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: DECREMENT {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "UCASE" Then
						If iArgCount = 0 Then       ' Act on content
							sContent = UCase (sContent)
						ElseIf iArgCount = 1 Then   ' Act on token
							If Not StoreToken (asArgs (1), UCase (GetTokenValue (asArgs (1), ""))) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: UCASE [{token_name}]"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "LCASE" Then
						If iArgCount = 0 Then       ' Act on content
							sContent = UCase (sContent)
						ElseIf iArgCount = 1 Then   ' Act on token
							If Not StoreToken (asArgs (1), LCase (GetTokenValue (asArgs (1), ""))) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: LCASE [{token_name}]"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "FOR" Then
						If UCase (asArgs (2)) = "=" AND UCase (asArgs (4)) = "TO" AND (iArgCount = 5 OR (iArgCount = 7 AND UCase (asArgs (6)) = "STEP")) Then
							iForNextDepth = iForNextDepth + 1
							aiForNextLines(iForNextDepth, 0) = iLineNum               ' 0 = line index
							aiForNextLines(iForNextDepth, 1) = asArgs (1)             ' 1 = token name
						  iErr = 0
							If iArgCount = 7 Then
							  iErr = 0
							  On Error Resume Next
								aiForNextLines(iForNextDepth, 2) = Cint (EVal (ReplaceTokens (asArgs (7), sTokenPattern, sNamePattern)))
								iErr = Err.Number
								sErr = Err.Description
								On Error Goto 0
								If iErr <> 0 Then
    							bFatal = True
 							  End If
							Else
								aiForNextLines(iForNextDepth, 2) = 1                  ' default increment
							End If
							If iErr = 0 Then
							  On Error Resume Next
							  aiForNextLines(iForNextDepth, 3) = CInt (EVal (ReplaceTokens (asArgs (5), sTokenPattern, sNamePattern)))
 								iErr = Err.Number
 								sErr = Err.Description
 								On Error Goto 0
 								If iErr <> 0 Then
     							bFatal = True
 								End If
							End If
							If iErr = 0 Then
							  iErr = 0
							  On Error Resume Next
							  nX = CInt (EVal (ReplaceTokens (asArgs (3), sTokenPattern, sNamePattern)))
								iErr = Err.Number
								sErr = Err.Description
								On Error Goto 0
								If iErr <> 0 Then
    							bFatal = True
								End If
							End If
							If iErr = 0 Then
  							If Not StoreToken (asArgs (1), CStr (nX)) Then
  								iErr = -1
  								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
  								bFatal = True
  							End If
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: FOR {token} = {initial} TO {test} [STEP {offset}]"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "NEXT" Then
						If iArgCount = 0 Then
							If iForNextDepth = -1 Then
								iErr = -1
								sErr = "NEXT without matching FOR"
								bFatal = True
							Else
								' Setup the next loop
								Call StoreToken ( _
									  aiForNextLines (iForNextDepth, 1), _
									  CStr (CInt ("0" & GetTokenValue (aiForNextLines (iForNextDepth, 1), "")) + aiForNextLines (iForNextDepth, 2)) _
									  )
								iIter = CInt (GetTokenValue (aiForNextLines (iForNextDepth, 1), ""))
								iTest = aiForNextLines (iForNextDepth, 3)
								If SGN (iIter - iTest) = SGN (aiForNextLines (iForNextDepth, 2)) Then
									' Loop has completed
									iForNextDepth = iForNextDepth - 1
								Else
									' Move pointer to the BEGIN_ITERATION line.
									' The increment before LOOP will move it to the statement following it.
									iLineNum = aiForNextLines(iForNextDepth, 0)
								End If
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: NEXT"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "EXTRACT_ELEMENTS" Then
						bTmp = (iArgCount = 9)
						If bTmp Then
							bTmp = (UCase (asArgs (2)) = "FROM" AND UCase (asArgs (6)) = "TO")
						End If

						If bTmp Then
							Err.Clear
							On Error Resume Next
							sTmp = RegExFramedFind _
										(sContent, _
										ReplaceTokens (asArgs (3), sTokenPattern, sNamePattern), _
										CBool (Eval (ReplaceTokens (asArgs (4), sTokenPattern, sNamePattern))), _
										CInt (ReplaceTokens (asArgs (5), sTokenPattern, sNamePattern)), _
										ReplaceTokens (asArgs (7), sTokenPattern, sNamePattern), _
										CBool (Eval (ReplaceTokens (asArgs (8), sTokenPattern, sNamePattern))), _
										CInt (ReplaceTokens (asArgs (9), sTokenPattern, sNamePattern)))
							iErr = Err.Number
							sErr = Err.Description
							On Error Goto 0
							bTmp = (iErr = 0)
							If Not bTmp Then
								bFatal = True
							End If

							If Not bFatal Then
								bTmp = StoreToken (asArgs(1) & ".COUNT", GetColumnFromString (sTmp, ",", 0))
								If Not bTmp Then
									iErr = -1
									sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
									bFatal = True
								End If
							End If

							If Not bFatal Then
								bTmp = StoreToken (asArgs(1) & ".LOCATIONS", sTmp)
								If Not bTmp Then
									iErr = -1
									sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
									bFatal = True
								End If
							End If

							If Not bFatal Then
								bTmp = StoreToken (asArgs(1) & ".SOURCE", sContent)
								If Not bTmp Then
									iErr = -1
									sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
									bFatal = True
								End If
							End If

							If Not bFatal Then
								If CInt (GetColumnFromString (GetTokenValue (asArgs (1) & ".LOCATIONS", ""), ",", 0)) = 0 Then
									sContent = ""
								Else
									sTmp = GetTokenValue (asArgs (1) & ".LOCATIONS", "")
									lTmp1 = CLng (GetColumnFromString (sTmp, "," ,1))
									lTmp2 = CLng (GetColumnFromString (sTmp, "," ,2))
									Err.Clear
									On Error Resume Next
									sContent = Mid (GetTokenValue (asArgs (1) & ".SOURCE", ""), lTmp1, lTmp2)
									iErr = Err.Number
									sErr = Err.Description
									On Error Goto 0
									bTmp = (iErr = 0)
									If Not bTmp Then
										bFatal = True
									End If
								End If
							End If

						Else
							iErr = -1
							sErr = "Does not conform to prototype: EXTRACT_ELEMENTS {token} FROM {regex1} {iteration} {ignore_case} TO {regex2} {iteration} {ignore_case}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "USE_ELEMENT" Then
						bTmp = (iArgCount = 2)
						If bTmp Then
							bTmp = TokenExists (asArgs (1) & ".COUNT")
							If Not bTmp Then
								iErr = -1
								sErr = "ELEMENT instruction references an undefined token name: " & asArgs(1)
								bFatal = True
							End If
						End If

						If bTmp Then
							lTmp = CInt (EVal (ReplaceTokens (asArgs (2), sTokenPattern, sNamePattern)))
							bTmp = (lTmp >= 0)
							If Not bTmp Then
								iErr = -1
								sErr = "ELEMENT position must be 0 or greater"
								bFatal = True
							End If
						End If

						If bTmp Then
							bTmp = (lTmp <= CInt (GetTokenValue (asArgs (1) & ".COUNT", "")))
							If Not bTmp Then
								iErr = -1
								sErr = "ELEMENT specifies an array match beyond the number of matches: " & GetTokenValue (asArgs (1) & ".COUNT", "")
								bFatal = True
							End If
						End If

						If bTmp Then
							sTmp = GetTokenValue (asArgs (1) & ".LOCATIONS", "")
							If CInt (GetColumnFromString (sTmp, ",", 0)) = 0 Then
								sContent = ""
							Else
								lTmp1 = CLng (GetColumnFromString (sTmp, "," ,lTmp * 2 + 1))
								lTmp2 = CLng (GetColumnFromString (sTmp, "," ,lTmp * 2 + 2))
								Err.Clear
								On Error Resume Next
								sContent = Mid (GetTokenValue (asArgs (1) & ".SOURCE", ""), lTmp1, lTmp2)
								iErr = Err.Number
								sErr = Err.Description
								On Error Goto 0
								bTmp = (iErr = 0)
								If Not bTmp Then
									bFatal = True
								End If
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: USE_ELEMENT {token} {index}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "EXTRACT" Then
						If iArgCount = 3 Then
							Err.Clear
							On Error Resume Next
							sContent = RegExFind ( _
                       ReplaceTokens (asArgs(1), sTokenPattern, sNamePattern), _
                       sContent, _
                       CBool(Eval (ReplaceTokens (asArgs (2), sTokenPattern, sNamePattern))), _
                       CInt(ReplaceTokens (asArgs (3), sTokenPattern, sNamePattern)))(0)
							iErr = Err.Number
							sErr = Err.Description
							On Error Goto 0
							If iErr <> 0 Then
							   bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: EXTRACT {regex} {ignore_case} {iteration}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "EVAL" Then
						If iArgCount >= 1 Then
							sTmp = ReplaceTokens (Mid (asLines (iLineNum), aiArgPos (1, 0)), sTokenPattern, sNamePattern)
							Err.Clear
							On Error Resume Next
							sContent = Eval (sTmp)
							iErr = Err.Number
							sErr = Err.Description
							On Error Goto 0
							sErrLine = "EVAL " & sTmp
						Else
							iErr = -1
							sErr = "Does not conform to prototype: EVAL {expression}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "STRIP" Then
						If iArgCount = 1 Then
							sTmp = ReplaceTokens (asArgs(1), sTokenPattern, sNamePattern)
							Err.Clear
							On Error Resume Next
							sContent = Replace (sContent, Chr (CByte (sTmp)), "")
							iErr = Err.Number
							sErr = Err.Description
							On Error Goto 0
							sErrLine = "STRIP " & sTmp
						Else
							iErr = -1
							sErr = "Does not conform to prototype: STRIP {ascii_value}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "REPLACE" Then
						If iArgCount = 2 Then
							sTmp1 = ReplaceTokens (asArgs(1), sTokenPattern, sNamePattern)
							sTmp2 = ReplaceTokens (asArgs(2), sTokenPattern, sNamePattern)
							Err.Clear
							On Error Resume Next
							sContent = Replace (sContent, sTmp1, sTmp2)
							iErr = Err.Number
							sErr = Err.Description
							On Error Goto 0
							sErrLine = "STRIP " & sTmp
						Else
							iErr = -1
							sErr = "Does not conform to prototype: REPLACE {search} {value}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "LITERAL" Then
						If iArgCount >= 1 Then
							sContent = Mid (asLines (iLineNum), aiArgPos (1, 0))
						Else
							iErr = -1
							sErr = "Does not conform to prototype: LITERAL {anything}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "SAVE" Then
						If iArgCount = 1 Then
							If Not StoreToken (asArgs (1), sContent) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							ElseIf Not StoreToken (asArgs (1) & ".LENGTH", LEN(sContent)) Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: SAVE {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "RECALL" Then
						If iArgCount = 1 Then
							If TokenExists (asArgs (1)) Then
								sContent = GetTokenValue (asArgs (1), "")
							Else
								iErr = -1
								sErr = "No such token exists"
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: RECALL {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf InStr (" STORE STORE_RAW STORELINE STORELINE_RAW APPEND APPEND_RAW APPENDLINE APPENDLINE_RAW ", " " & asArgs (0) & " ") > 0 Then
						If iArgCount >= 1 Then
							If iArgCount >= 2 Then
								If InStr (asArgs (0), "_RAW") > 0 Then
									sTmp = Mid (asLines (iLineNum), aiArgPos (2, 0))
								Else
									sTmp = ReplaceTokens (Mid (asLines (iLineNum), aiArgPos (2, 0)), sTokenPattern, sNamePattern)
								End If
							Else
								sTmp = ""
							End If
							asArgs (0) = Replace (asArgs (0), "_RAW", "")
							If asArgs (0) = "APPEND"  Or asArgs (0) = "APPENDLINE" Then
								If Not TokenExists (asArgs (1)) Then
									bResult = StoreToken (asArgs (1), "")
								End If
							End If

							If asArgs (0) = "STORE" Then bResult = StoreToken (asArgs (1), sTmp)
							If asArgs (0) = "STORE" Then bResult = StoreToken (asArgs (1) & ".LENGTH", LEN(sTmp))
							If asArgs (0) = "STORELINE" Then bResult = StoreToken (asArgs (1), sTmp & vbCrLf)
							If asArgs (0) = "STORELINE" Then bResult = StoreToken (asArgs (1) & ".LENGTH", LEN(sTmp & vbCrLf))
							If asArgs (0) = "APPEND" Then bResult = StoreToken (asArgs (1), GetTokenValue (asArgs (1), "") & sTmp)
							If asArgs (0) = "APPEND" Then bResult = StoreToken (asArgs (1) & ".LENGTH", LEN (GetTokenValue (asArgs (1), "")))
							If asArgs (0) = "APPENDLINE" Then bResult = StoreToken (asArgs (1), GetTokenValue (asArgs (1), "") & sTmp & vbCrLf)
							If asArgs (0) = "APPENDLINE" Then bResult = StoreToken (asArgs (1) & ".LENGTH", LEN (GetTokenValue (asArgs (1), "") & vbCrLf))

							If Not bResult Then
								iErr = -1
								sErr = "Available token storage is exhausted (maximum " & UBound (asTokens, 2) & ")"
								bFatal = True
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: " & asArgs (0) & " {token_name} {text}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf (asArgs (0) = "OUTPUT" Or asArgs (0) = "OUTPUTLINE") Then
						If iArgCount >= 1 OR asArgs (0) = "OUTPUTLINE" Then
							If iArgCount >= 1 Then
								sTmp = ReplaceTokens (Mid (asLines (iLineNum), aiArgPos (1, 0)), sTokenPattern, sNamePattern)
							Else
								sTmp = ""
							End If
							If asArgs (0) = "OUTPUT" Then oOutput.WriteText sTmp, 0
							If asArgs (0) = "OUTPUTLINE" Then oOutput.WriteText sTmp, 1
						Else
							iErr = -1
							sErr = "Does not conform to prototype: " & asArgs (0) & " {text}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "DROP" Then
						If iArgCount = 1 Then
							DropToken (asArgs (1))
						Else
							iErr = -1
							sErr = "Does not conform to prototype: DROP {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "RETURN" Then
						If iArgCount = 1 Then
							If Not TokenExists (asArgs (1)) Then
								iErr = -1
								sErr = "No such token exists"
								bFatal = True
							Else
								sOutput = GetTokenValue (asArgs (1), "")
								bUsingReturn = True
								Exit Do
							End If
						Else
							iErr = -1
							sErr = "Does not conform to prototype: RETURN {token_name}"
							bFatal = True
						End If

					' -----------------------------------------------------------------------
					ElseIf asArgs (0) = "RAISE_ERROR" Then
						iErr = -1
						sErr = "Script intentionally aborted"
						bFatal = True

					' -----------------------------------------------------------------------
					Else
							iErr = -1
							sErr = "Unknown command token"
							bFatal = True
					End If
				End If
			End If

			If iErr <> 0 Then
				Call StoreToken ("__ERROR_NUM__", CStr (iErr))
				Call StoreToken ("__ERROR_MSG__", sErr)
				Call StoreToken ("__ERROR_LINE__", LTrim (asLines (iLineNum)))
				Call StoreToken ("__ERROR_LINENUM__", iLineNum + 1)

     			sErrors = sErrors & ReplaceTokens (GetTokenValue ("__ERROR_BODY__", ""), sTokenPattern, sNamePattern)
			End If
			iLineNum = iLineNum + 1

		Loop

		If iIfBlockLevel > 0 And Not bFatal And Not bUsingReturn Then
			iErr = -1
			sErr = "Orphan IF statement: no matching ENDIF"
			Call StoreToken ("__ERROR_NUM__", CStr (iErr))
			Call StoreToken ("__ERROR_MSG__", sErr)
			Call StoreToken ("__ERROR_LINE__", LTrim (asLines (iLineNum)))
			Call StoreToken ("__ERROR_LINENUM__", iLineNum + 1)

   		sErrors = sErrors & ReplaceTokens (GetTokenValue ("__ERROR_BODY__", ""), sTokenPattern, sNamePattern)
		End If

		If sErrors <> "" And Not bUsingReturn Then
			sErrors = GetTokenValue ("__ERROR_PREFIX__", "") & sErrors & GetTokenValue ("__ERROR_SUFFIX__", "")
		End If

		If sOutput = "" And Not bUsingReturn Then
			With oOutput
				.Position = 0
				sOutput = .ReadText()
			End With
		End If
		oOutput.Close()
		set oOutput = Nothing

		If sErrors <> "" And Not bUsingReturn Then
			If CBool (Eval (GetTokenValue ("__ERRORS_ONLY__", ""))) Then
				sOutput = sErrors
			Else
				sOutput = sErrors & sOutput
			End If
		End If

		If Not bUsingReturn Then
			sOutput = _
				ReplaceTokens (GetTokenValue ("__OUTPUT_HEADER__", ""), sTokenPattern, sNamePattern) & _
				sOutput & _
				ReplaceTokens (GetTokenValue ("__OUTPUT_FOOTER__", ""), sTokenPattern, sNamePattern)
		End If

		ScrapeUsingScript = sOutput

	End Function

End Class

